from dataclasses import dataclass
from flippy.core import BaseBackend, FeatureName, Gate
from flippy.exceptions import FeatureNotFound

ACTOR_IF_NO_TARGET = "anonymous"


@dataclass
class FeatureState:
    key: str
    boolean: bool | None
    actors: list[str]
    groups: list[str]
    percent_actors: int | None
    percent_time: int | None


class Flippy:
    def __init__(self, backend: BaseBackend):
        self._backend = backend
    
    def is_enabled(self, feature: FeatureName, target = None) -> bool:
        try:
            f = self._backend.get(feature)
        except FeatureNotFound:
            return False

        # if the boolean gate is on or off, that's final
        match f.state:
            case 'on':
                return True
            case 'off':
                return False

        # if the feature is conditional and no target was given, use a constant
        if target is None:
            actor = ACTOR_IF_NO_TARGET
        else:
            actor = self._to_flipper_id(target)

        if any([
            f.actors_gate.is_open(actor, feature),
            f.groups_gate.is_open(actor, feature),
            f.percentage_of_actors_gate.is_open(actor, feature),
            f.percentage_of_time_gate.is_open(actor, feature),
            f.expression_gate.is_open(actor, feature),
        ]):
            return True
        
        # TODO: special check whether actor is a member of an enabled group
        # (using the Django authentication system)
        ...
        
        return False
    
    def create(self, feature: FeatureName) -> bool:
        return self._backend.add(feature)
    
    def get_all_feature_names(self) -> set[FeatureName]:
        return {f.key for f in self._backend.get_all()}
    
    def feature_exists(self, feature: FeatureName) -> bool:
        try:
            self._backend.get(feature)
            return True
        except FeatureNotFound:
            return False
    
    def get_feature_state(self, feature: FeatureName) -> FeatureState:
        f = self._backend.get(feature)
        return FeatureState(
            key=f.key,
            boolean=f.boolean_gate.value,
            actors=f.actors_gate.value,
            groups=f.groups_gate.value,
            percent_actors=f.percentage_of_actors_gate.value,
            percent_time=f.percentage_of_time_gate.value,
        )
    
    def enable(self, feature: FeatureName) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Boolean)
        except FeatureNotFound:
            pass

    def enable_actor(self, feature: FeatureName, target) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Actors, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def enable_group(self, feature: FeatureName, target) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Groups, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def enable_percentage_of_actors(self, feature: FeatureName, percentage: int) -> None:
        assert isinstance(percentage, int)
        assert 0 <= percentage <= 100

        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.PercentageOfActors, percentage)
        except FeatureNotFound:
            pass

    def enable_percentage_of_time(self, feature: FeatureName, percentage: int) -> None:
        assert isinstance(percentage, int)
        assert 0 <= percentage <= 100

        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.PercentageOfTime, percentage)
        except FeatureNotFound:
            pass

    def disable(self, feature: FeatureName) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Boolean)
        except FeatureNotFound:
            pass

    def disable_actor(self, feature: FeatureName, target) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Actors, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def disable_group(self, feature: FeatureName, target) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Groups, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def disable_percentage_of_actors(self, feature: FeatureName) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.PercentageOfActors)
        except FeatureNotFound:
            pass

    def disable_percentage_of_time(self, feature: FeatureName) -> None:
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.PercentageOfTime)
        except FeatureNotFound:
            pass
    
    def clear(self, feature: FeatureName) -> bool:
        return self._backend.clear(feature)

    def destroy(self, feature: FeatureName) -> bool:
        return self._backend.remove(feature)

    def _to_flipper_id(self, object) -> str:
        if hasattr(object, 'get_flipper_id'):
            return object.get_flipper_id()

        type_name = object.__class__.__name__
        object_id = None

        if hasattr(object, 'flipper_id'):
            object_id = object.flipper_id
        elif hasattr(object, 'pk'):
            object_id = object.pk
        elif hasattr(object, 'id'):
            object_id = object.id
        else:
            # TODO: check if this is stable over time
            object_id = hash(object)
        
        return f"{type_name};{object_id}"



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Literal, NewType

from flippy.gates import (ActorsGate, BooleanGate, ExpressionGate, GroupsGate,
                          PercentageOfActorsGate, PercentageOfTimeGate)

FeatureName = NewType('FeatureName', str)

FlagState = Literal['on', 'conditional', 'off']


class Gate(Enum):
    Boolean = 'boolean'
    Actors = 'actors'
    Groups = 'groups'
    PercentageOfActors = 'percentage_of_actors'
    PercentageOfTime = 'percentage_of_time'
    Expression = 'expression'


@dataclass
class Feature:
    key: FeatureName

    @property
    def state(self) -> FlagState:
        if self.boolean_gate.value:
            return 'on'
        if (self.actors_gate.value == []
            and self.groups_gate.value == []
            and self.percentage_of_actors_gate.value == None
            and self.percentage_of_time_gate.value == None
            and self.expression_gate.value == None):
            return 'off'
        return 'conditional'

    boolean_gate: BooleanGate = field(default_factory=BooleanGate)
    actors_gate: ActorsGate = field(default_factory=ActorsGate)
    groups_gate: GroupsGate = field(default_factory=GroupsGate)
    percentage_of_actors_gate: PercentageOfActorsGate = field(default_factory=PercentageOfActorsGate)
    percentage_of_time_gate: PercentageOfTimeGate = field(default_factory=PercentageOfTimeGate)
    expression_gate: ExpressionGate = field(default_factory=ExpressionGate)

    def to_api(self):
        return {
            'key': self.key,
            'state': self.state,
            'gates': [
                self.boolean_gate.to_api(),
                self.actors_gate.to_api(),
                self.groups_gate.to_api(),
                self.percentage_of_actors_gate.to_api(),
                self.percentage_of_time_gate.to_api(),
                self.expression_gate.to_api(),
            ]
        }
    
    @classmethod
    def from_api(cls, api_payload: dict):
        self = cls(api_payload['key'])
        for gate in api_payload['gates']:
            gateValue = gate['value']
            match gate['key']:
                case 'boolean':
                    self.boolean_gate.value = gateValue
                case 'actors':
                    self.actors_gate.value = gateValue
                case 'groups':
                    self.groups_gate.value = gateValue
                case 'percentage_of_actors':
                    self.percentage_of_actors_gate.value = None if gateValue is None else int(gateValue)
                case 'percentage_of_time':
                    self.percentage_of_time_gate.value = None if gateValue is None else int(gateValue)
                case 'expression':
                    self.expression_gate.value = gateValue
                case _:
                    raise ValueError(f"{gate} is not a known gate type")
        return self


class BaseBackend(metaclass=ABCMeta):
    """
    Flipper established a protocol for adapters to implement.
    This is the Pythonic implementation of it.
    
    https://www.flippercloud.io/docs/adapters/new
    """
    @abstractmethod
    def features(self) -> set[FeatureName]:
        "Get the set of known features."
        pass

    @abstractmethod
    def add(self, feature: FeatureName) -> bool:
        "Add a feature to the set of known features."
        pass

    @abstractmethod
    def remove(self, feature: FeatureName) -> bool:
        "Remove a feature from the set of known features."
        pass

    @abstractmethod
    def clear(self, feature: FeatureName) -> bool:
        "Clear all gate values for a feature."
        pass

    @abstractmethod
    def get(self, feature: FeatureName) -> Feature:
        "Get all gate values for a feature."
        pass

    @abstractmethod
    def enable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Enable a gate for a thing."
        pass

    @abstractmethod
    def disable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Disable a gate for a thing."
        pass

    @abstractmethod
    def get_multi(self, features: list[FeatureName]) -> list[Feature]:
        "Get all gate values for several features at once."
        # default implementation; feel free to use or override
        values = []
        for feature in features:
            values.append(self.get(feature))
        return values

    @abstractmethod
    def get_all(self) -> list[Feature]:
        "Get all gate values for all features at once."
        # default implementation; feel free to use or override
        values = []
        for feature in self.features():
            values.append(self.get(feature))
        return values

    # dict implementation
    # note, there's no setting of values because that doesn't fit the
    # dict contract cleanly
    def __len__(self):
        return len(self.features())
    
    def __getitem__(self, key: FeatureName):
        return self.get(key)
    
    def __delitem__(self, key: FeatureName):
        return self.remove(key)
    
    def __iter__(self):
        return iter(self.features())
    
    def __contains__(self, key: FeatureName):
        return key in self.features()
    # end dict implementation

BaseBackend.register(dict)



from flippy.backends.flipper_cloud import FlipperCloudBackend
from flippy.backends.memory import MemoryBackend
from flippy.backends.django import DjangoBackend

__all__ = [
    FlipperCloudBackend,
    MemoryBackend,
    DjangoBackend,
]
